/*
 * Copyright (c) 2024 Google LLC.
 *
 * SPDX-License-Identifier: Apache-2.0
 */

#include <zephyr/drivers/gpio.h>
#include <zephyr/kernel.h>

#include <ch32_gpio.h>
#include <ch32_uart.h>

namespace
{

constexpr struct gpio_dt_spec kLed = GPIO_DT_SPEC_GET(DT_ALIAS(led0), gpios);
constexpr struct gpio_dt_spec kScl = GPIO_DT_SPEC_GET(DT_ALIAS(scl), gpios);
constexpr struct gpio_dt_spec kSda = GPIO_DT_SPEC_GET(DT_ALIAS(sda), gpios);

void sniff()
{
	GPIO_TypeDef *port = GPIOC;
	USART_TypeDef *uart = USART1;
	uint8_t fifo[256];
	int head = 0;
	int tail = 0;

	// Note that using `int` reduces the code size by removing unneeded truncation.
	int samples = 0;
	int sample_count = 0;
	int last_sample = 0;
	int unchanged_for = 0;

	csr_clear(mstatus, MSTATUS_IEN);

	for (;;) {
		static_assert(kScl.port == kSda.port);
		static_assert(kScl.pin == kSda.pin + 1);
		int sample = (port->INDR >> kSda.pin) & 3;

		// Record when the value changes, or the value has been unchanged for some time.
		if (sample != last_sample || ++unchanged_for >= 1024) {
			samples <<= 2;
			samples |= sample;
			last_sample = sample;
			unchanged_for = 0;

			if (++sample_count >= 3) {
				// Encode the three samples into valid ASCII.
				fifo[head++] = samples + ' ';
				head %= sizeof(fifo);
				samples = 0;
				sample_count = 0;
			}
		}
		if (tail != head && (uart->STATR & USART_STATR_TXE) != 0) {
			uart->DATAR = fifo[tail++];
			tail %= sizeof(fifo);

			if (tail == 0 && (uart->STATR & USART_STATR_RXNE) != 0 &&
			    uart->DATAR == '~') {
				break;
			}
		}
	}

	csr_set(mstatus, MSTATUS_IEN);
}

void interceptor()
{
	gpio_pin_configure_dt(&kLed, GPIO_OUTPUT_ACTIVE);
	gpio_pin_configure_dt(&kScl, GPIO_INPUT);
	gpio_pin_configure_dt(&kSda, GPIO_INPUT);

	for (;;) {
		printk("sniff\n");
		sniff();
	}
}

} // namespace

int main(void)
{
	interceptor();
	return 0;
}
